#!/usr/bin/env python3

"""
The content of this file includes portions of the AUDIOKINETIC Wwise Technology
released in source code form as part of the SDK installer package.

Commercial License Usage

Licensees holding valid commercial licenses to the AUDIOKINETIC Wwise Technology
may use this file in accordance with the end user license agreement provided
with the software or, alternatively, in accordance with the terms contained in a
written agreement between you and Audiokinetic Inc.

  Copyright (c) 2021 Audiokinetic Inc.
"""

import argparse
import os
import re
import subprocess
import sys
import os.path as osp
from common.constant import PLUGIN_NAME, PREMAKE, PROJECT_ROOT, WWISE_ROOT
from common.platform import *
from common.registry import platform_registry, get_supported_platforms, is_authoring_target

SUPPORTED_PLATFORMS = get_supported_platforms("premake")

def run(argv):
    # parse the command line
    parser = argparse.ArgumentParser(description="Audiokinetic Inc. premake tool for plugins")
    parser.add_argument("platform", metavar="platform", choices=SUPPORTED_PLATFORMS, help="platform to premake ({})".format(", ".join(SUPPORTED_PLATFORMS)))
    parser.add_argument("--debugger", help="Enable lua debugger for premake scripts", action="store_true")
    args = parser.parse_args(argv)

    # prepare premake parameters
    platform_info = platform_registry.get(args.platform)
    pm_actions = platform_info.premake.actions

    pm_scripts = "--scripts={};{};{}".format(
        os.path.join(os.getenv('WWISESDK'), "source", "Build"),
        os.path.join(WWISE_ROOT, "Scripts", "Premake"),
        os.path.join(WWISE_ROOT, "Scripts", "Build"))

    pm_file = "--file={}".format(
        os.path.join(osp.dirname(__file__), "premakePlugins.lua"))

    # Default to PlatformInfo name if not specified
    pm_platform = "--os={}".format(
        platform_info.premake.platform or
        re.search(r"\w+(?=(_))|\w+", platform_info.name).group(0).lower()
    )

    pm_plugin_dir = "--plugindir={}".format(PROJECT_ROOT)

    pm_is_authoring = "--authoring={}".format("yes" if is_authoring_target(platform_info.name) else "no")

    # run premake
    for pm_action in pm_actions:
        cmd = [PREMAKE, pm_scripts, pm_file, pm_action, pm_platform, pm_plugin_dir, pm_is_authoring]
        if args.debugger:
            cmd.append("--debugger")

        print("Premake Command: " + str(cmd))
        res = subprocess.Popen(cmd).wait()
        if res != 0:
            sys.exit(res)

if __name__ == "__main__":
    run(sys.argv[1:])



#!/usr/bin/env python3

"""
The content of this file includes portions of the AUDIOKINETIC Wwise Technology
released in source code form as part of the SDK installer package.

Commercial License Usage

Licensees holding valid commercial licenses to the AUDIOKINETIC Wwise Technology
may use this file in accordance with the end user license agreement provided
with the software or, alternatively, in accordance with the terms contained in a
written agreement between you and Audiokinetic Inc.

  Copyright (c) 2021 Audiokinetic Inc.
"""

import argparse
import os
import platform
import sys
import re
from common.constant import PLUGIN_NAME, PROJECT_ROOT
from common.platform import *
# [wp-enhanced patch] import android platform patch
from wpe.wp_patch.common.platform import android
# [/wp-enhanced patch]
from common.registry import platform_registry, get_supported_platforms
from common.util import exit_with_error

SUPPORTED_PLATFORMS = get_supported_platforms("build", platform.system())

def run(argv):
    # parse the command line
    parser = argparse.ArgumentParser(description="Audiokinetic Inc. build tool for plugins")
    parser.add_argument("platform", metavar="platform", choices=SUPPORTED_PLATFORMS, help="platform to build ({})".format(", ".join(SUPPORTED_PLATFORMS)))
    parser.add_argument("-c", "--configuration", help="configuration to build (Debug, Release, Profile, ...).")
    parser.add_argument("-x", "--archs", default=[], nargs='+', help="architecture to build (x32, x64, ...).")
    parser.add_argument("-t", "--toolset", help="toolset used to build on Windows platforms (vc150, vc160, vc170).")
    parser.add_argument("-f", "--build-hooks-file", help="path to a Python file defining one or more of the supported hooks (postbuild) to be called at various step during the build process")
    parser.add_argument("--toolchain-vers", help="Path to a \'ToolchainVers\' text file, containing a list of supported toolchain versions to pass to the platform's toolchain_setup script, to setup and define a set of env-vars to re-run each build step with.")
    parser.add_argument("--toolchain-env-script", help="Path to a \'GetToolchainEnv\' script, which, when executed with a version provided by the toolchain-vers file, returns a comma separated list of environment variables to apply for a build step.")
    args = parser.parse_args(argv)

    # import the build hooks
    build_hooks = None
    if args.build_hooks_file:
        sys.path.append(os.path.join(PROJECT_ROOT))
        if not os.path.exists(args.build_hooks_file):
            exit_with_error("Missing build hooks file {} at {}".format(args.build_hooks_file, PROJECT_ROOT))
        try:
            build_hooks = __import__(os.path.splitext(args.build_hooks_file)[0])
        except Exception as e:
            exit_with_error("Invalid build hooks file {} at {}\n{}".format(args.build_hooks_file, PROJECT_ROOT, str(e)))

    # handle platform-specific command line specificites
    platform_info = platform_registry.get(args.platform)

    invalid_archs = set(args.archs) - set(platform_info.build.archs)
    if invalid_archs:
        print("error: Invalid architecture(s) '{}' for target {}.".format(invalid_archs, platform_info.name))
        print("supported architectures: {}".format(', '.join(platform_info.build.archs)))
        return 1

    if not args.toolset:
        if platform_info.build.require_toolset:
            print("error: 'Toolset' argument for target {} needs to be specified using -t or --toolset, -h for more details.".format(platform_info.name))
            print("supported toolsets: {}".format(', '.join(platform_info.build.toolsets)))
            return 1
    elif args.toolset not in platform_info.build.toolsets:
        print("error: Invalid toolset '{}' for target {}.".format(args.toolset, platform_info.name))
        print("supported toolsets: {}".format(', '.join(platform_info.build.toolsets)))
        return 1

    if not args.configuration:
        if platform_info.build.require_configuration:
            exit_with_error("'Configuration' argument for target {} needs to be specified using -c or --configuration, -h for more details.".format(platform_info.name))

        print("Building {} for {}.".format(PLUGIN_NAME, platform_info.name))
    else:
        if args.configuration not in platform_info.build.configurations:
            print("error: Invalid configuration '{}' for target {}.".format(args.configuration, platform_info.name))
            print("supported configurations: {}".format(', '.join(platform_info.build.configurations)))
            return 1

        print("Building {} for {} in {}...".format(PLUGIN_NAME, platform_info.name, args.configuration))

    if args.archs:
        # build for the requested architecture instead of building them all
        platform_info.build.archs = args.archs

    if args.toolset:
        # build for the requested toolset instead of building them all
        platform_info.build.toolsets = (args.toolset,)

    if args.toolchain_vers:
        # specify an override location for the toolchain vers file
        platform_info.build.toolchain_vers = args.toolchain_vers

    if args.toolchain_env_script:
        # specify an override location for the toolchain env script
        platform_info.build.toolchain_env_script = args.toolchain_env_script

    # run the build
    return platform_info.build.command(platform_info.name, args.configuration, build_hooks)

if __name__ == "__main__":
    sys.exit(run(sys.argv[1:]))




extra_gitignore = '''
# premake-androidmk
*.mk
.xcode_development_team


# wp output logs
Output

external'''


rider_run_config = r'''
    <configuration name="run with wwise" type="CppProject" factoryName="C++ Project">
      <configuration_1>
        <option name="CONFIGURATION" value="Debug" />
        <option name="PLATFORM" value="x64" />
        <option name="PROJECT_FILE_PATH" value="$PROJECT_DIR$/WwisePlugin/MiaBinauralizer_Authoring_Windows_vc160.vcxproj" />
        <option name="CURRENT_LAUNCH_PROFILE" value="Local" />
        <option name="EXE_PATH" value="$(WWISEROOT)\Authoring\x64\Release\bin\Wwise.exe" />
        <option name="PROGRAM_PARAMETERS" value="&quot;D:\UE4TestProjs\PluginPlaygroundUE\PluginPlaygroundUE_WwiseProject\PluginPlaygroundUE_WwiseProject.wproj&quot;" />
        <option name="WORKING_DIRECTORY" value="$(ProjectDir)" />
        <option name="PASS_PARENT_ENVS" value="1" />
        <option name="USE_EXTERNAL_CONSOLE" value="0" />
      </configuration_1>
      <configuration_2>
        <option name="CONFIGURATION" value="Release" />
        <option name="PLATFORM" value="x64" />
        <option name="PROJECT_FILE_PATH" value="$PROJECT_DIR$/WwisePlugin/MiaBinauralizer_Authoring_Windows_vc160.vcxproj" />
        <option name="CURRENT_LAUNCH_PROFILE" value="Local" />
        <option name="EXE_PATH" value="$(WWISEROOT)\Authoring\x64\Release\bin\Wwise.exe" />
        <option name="PROGRAM_PARAMETERS" value="&quot;D:\UE4TestProjs\PluginPlaygroundUE\PluginPlaygroundUE_WwiseProject\PluginPlaygroundUE_WwiseProject.wproj&quot;" />
        <option name="WORKING_DIRECTORY" value="$(ProjectDir)" />
        <option name="PASS_PARENT_ENVS" value="1" />
        <option name="USE_EXTERNAL_CONSOLE" value="0" />
      </configuration_2>
      <configuration_3>
        <option name="CONFIGURATION" value="Debug" />
        <option name="PLATFORM" value="Win32" />
        <option name="PROJECT_FILE_PATH" value="$PROJECT_DIR$/WwisePlugin/MiaBinauralizer_Authoring_Windows_vc160.vcxproj" />
        <option name="CURRENT_LAUNCH_PROFILE" value="Local" />
        <option name="EXE_PATH" value="$(WWISEROOT)\Authoring\x64\Release\bin\Wwise.exe" />
        <option name="PROGRAM_PARAMETERS" value="&quot;D:\WwiseTestProjs\2021_3_plugin_test\2021_3_plugin_test.wproj&quot;" />
        <option name="WORKING_DIRECTORY" value="$(ProjectDir)" />
        <option name="PASS_PARENT_ENVS" value="1" />
        <option name="USE_EXTERNAL_CONSOLE" value="0" />
      </configuration_3>
      <configuration_4>
        <option name="CONFIGURATION" value="Release" />
        <option name="PLATFORM" value="Win32" />
        <option name="PROJECT_FILE_PATH" value="$PROJECT_DIR$/WwisePlugin/MiaBinauralizer_Authoring_Windows_vc160.vcxproj" />
        <option name="CURRENT_LAUNCH_PROFILE" value="Local" />
        <option name="EXE_PATH" value="$(WWISEROOT)\Authoring\x64\Release\bin\Wwise.exe" />
        <option name="PROGRAM_PARAMETERS" value="&quot;D:\WwiseTestProjs\2021_3_plugin_test\2021_3_plugin_test.wproj&quot;" />
        <option name="WORKING_DIRECTORY" value="$(ProjectDir)" />
        <option name="PASS_PARENT_ENVS" value="1" />
        <option name="USE_EXTERNAL_CONSOLE" value="0" />
      </configuration_4>
      <option name="DEFAULT_PROJECT_PATH" value="" />
      <method v="2" />
    </configuration>'''


import copy
from typing import Any, Optional
from dataclasses import dataclass, field
import xml.etree.ElementTree as ET
from collections import OrderedDict


# project
from wpe.util import *
from wpe.project_config import ProjectConfig, PluginInfo

_type_prefix_map = {
    'float': 'f',
    'int': 'i',
    'uint': 'u',
    'bool': 'b',
}

_wwise_type_name_map = {
    'float': 'AkReal32',
    'int': 'AkInt32',
    'uint': 'AkUInt32',
    'bool': 'bool',
}

_xml_type_name_map = {
    'float': 'Real32',
    'int': 'int32',
    'uint': 'Uint32',
    'bool': 'bool',
}

_supported_rtpc_types = {'Additive', 'Multiplicative', 'Exclusive', 'Boolean'}


@dataclass
class InnerType:
    name: str
    fields: list['Parameter']

    def __post_init__(self):
        self.structName = util.convert_compound_cases(self.name)

    @staticmethod
    def create(name, dict_define: dict[str, Any]):
        instance = InnerType(
            name=name,
            fields=[Parameter.create(name, defines) for name, defines in dict_define.items()],
        )
        return instance

    def generate_struct_defines(self):
        return f'''struct {self.structName}
{{
    {self.__generate_field_lines()}
}};'''

    def __generate_field_lines(self):
        lines = []
        for type_field in self.fields:
            type_field.generate_names()
            lines.append(type_field.generate_declaration())
        return '\n    '.join(lines)


@dataclass
class Parameter:
    name: str
    type_: str
    rtpc_type: str
    defaultValue: Any
    minValue: Any
    maxValue: Any
    data_meaning: str = ''
    description: list[dict] = field(default_factory=list)
    dependencies: list[dict] = field(default_factory=list)
    displayName: str = ''
    enumeration: list[dict] = field(default_factory=list)
    userInterface: dict = field(default_factory=dict)
    id: int = 0
    parent: Optional[InnerType] = None
    basename: str = ''
    suffix: str = ''

    def __post_init__(self):
        if self.type_ not in _wwise_type_name_map:
            raise ValueError(f'Unsupported type: {self.type_} in parameter "{self.name}". Expected one of {_wwise_type_name_map.keys()}.')

        if self.rtpc_type and self.rtpc_type not in _supported_rtpc_types:
            raise ValueError(f'Unsupported rtpc type: {self.rtpc_type} in parameter "{self.name}". Expected one of {_supported_rtpc_types}.')

    def generate_names(self):
        if self.type_ == 'bool':
            self.defaultValue = str(self.defaultValue).lower()
            self.minValue = 'false'
            self.maxValue = 'true'
        self.propertyName = util.convert_compound_cases(self.name)
        self.cppVariableName = _type_prefix_map[self.type_] + util.convert_compound_cases(self.basename or self.propertyName)
        self.paramIDName = f'PARAM_{self.name.upper()}_ID'
        self.typeName = _wwise_type_name_map[self.type_]
        self.xmlTypeName = _xml_type_name_map[self.type_]
        self.struct = 'InnerType' if self.parent else ('RTPC' if self.rtpc_type else 'NonRTPC')
        self.displayName = self.displayName or util.convert_compound_cases(self.name, style='title')
        self.nameSpace = f'{self.struct}.{self.parent.name}{self.suffix}' if self.parent else self.struct

    def assign_id(self, _id: int):
        self.id = _id

    @staticmethod
    def create(name, dict_define: dict[str, Any]):
        return Parameter(
            name=name,
            type_=dict_define['type'],
            rtpc_type=dict_define.get('rtpc_type', 'Exclusive' if dict_define.get('rtpc') else ''),
            defaultValue=dict_define['default_value'],
            minValue=dict_define.get('min_value', None),
            maxValue=dict_define.get('max_value', None),
            data_meaning=dict_define.get('data_meaning', ''),
            description=dict_define.get('description', []),
            dependencies=copy.deepcopy(dict_define.get('dependencies', [])),
            displayName=dict_define.get('display_name', ''),
            enumeration=dict_define.get('enumeration', []),
            userInterface=dict_define.get('user_interface', '')
        )

    def generate_param_id(self) -> str:
        return f'static constexpr AkPluginParamID {self.paramIDName} = {self.id};'

    def generate_declaration(self) -> str:
        if self.parent:
            return f'{self.parent.structName} {self.parent.name}{self.suffix};'
        return f'{self.typeName} {self.cppVariableName};'

    def generate_init(self) -> str:
        return f'{self.nameSpace}.{self.cppVariableName} = {self.defaultValue};'

    def generate_read_bank_data(self) -> str:
        return f'{self.nameSpace}.{self.cppVariableName} = READBANKDATA({self.typeName}, pParamsBlock, in_ulBlockSize);'

    def generate_set_parameter(self) -> str:
        need_reinterpret = self.typeName != 'AkReal32' and bool(self.rtpc_type)
        interpret_pointer = f'static_cast<{self.typeName}>(*(AkReal32*)in_pValue)' if need_reinterpret else f'*(({self.typeName}*)in_pValue)'
        return f'''    case {self.paramIDName}:
        {self.nameSpace}.{self.cppVariableName} = {interpret_pointer};
        m_paramChangeHandler.SetParamChange({self.paramIDName});
        break;'''

    def generate_property_name_declaration(self) -> str:
        return f'extern const char* const sz{self.propertyName};'

    def generate_property_name_definition(self) -> str:
        return f'const char* const sz{self.propertyName} = "{self.propertyName}";'

    def generate_write_bank_data(self) -> str:
        writer = 'Write' + util.convert_compound_cases(self.typeName.lstrip('Ak'))
        getter = 'Get' + util.convert_compound_cases(self.typeName.lstrip('Ak'))
        return f'in_dataWriter.{writer}(m_propertySet.{getter}(in_guidPlatform, sz{self.propertyName}));'

    def generate_parameter_gui(self) -> list[str]:
        def _generate_dependencies():
            if not self.dependencies:
                return ''
            dependencies_element = ET.Element('Dependencies')
            for dep in self.dependencies:
                property_dep = ET.SubElement(dependencies_element, 'PropertyDependency')
                property_dep.attrib['Name'] = dep['obj'].propertyName
                property_dep.attrib['Action'] = 'Enable'
                condition = ET.SubElement(property_dep, 'Condition')
                if (condition_value := dep['condition']) == 'Enumeration':
                    enumeration = ET.SubElement(condition, condition_value)
                    enumeration.attrib['Type'] = dep['obj'].xmlTypeName
                    for value in dep['values']:
                        value_tag = ET.SubElement(enumeration, 'Value')
                        value_tag.text = str(value)
                if (condition_value := dep['condition']) == 'Range':
                    range_ = ET.SubElement(condition, condition_value)
                    range_.attrib['Type'] = dep['obj'].xmlTypeName
                    min_value = ET.SubElement(range_, 'Min')
                    min_value.text = str(dep['min'])
                    max_value = ET.SubElement(range_, 'Max')
                    max_value.text = str(dep['max'])
            ET.indent(dependencies_element)
            return '\n' + ET.tostring(dependencies_element).decode(util.TXT_CODEC)
        support_rtpc_type = f'SupportRTPCType="{self.rtpc_type}"' if self.rtpc_type else ''
        data_meaning = f'DataMeaning="{self.data_meaning}"' if self.data_meaning else ''

        def _generate_bool_gui_lines():
            return f'''<Property Name="{self.propertyName}" Type="{self.xmlTypeName}" {support_rtpc_type} {data_meaning} DisplayName="{self.displayName}">
  <DefaultValue>{self.defaultValue}</DefaultValue>
  <AudioEnginePropertyID>{self.id}</AudioEnginePropertyID>{_generate_dependencies()}
</Property>'''.splitlines()

        def _generate_int_gui_lines():
            user_interface = self.userInterface
            if not self.enumeration:
                return f'''<Property Name="{self.propertyName}" Type="{self.xmlTypeName}" {support_rtpc_type} {data_meaning} DisplayName="{self.displayName}">
                  <UserInterface {user_interface} />
                  <DefaultValue>{self.defaultValue}</DefaultValue>
                  <AudioEnginePropertyID>{self.id}</AudioEnginePropertyID>
                </Property>'''.splitlines()
            options = '\n        '.join(['<Value DisplayName="{}">{}</Value>'.format(opt['displayName'], opt['value']) for opt in self.enumeration])
            return f'''<Property Name="{self.propertyName}" Type="{self.xmlTypeName}" {support_rtpc_type} {data_meaning} DisplayName="{self.displayName}">
  <UserInterface {user_interface} />
  <DefaultValue>{self.defaultValue}</DefaultValue>
  <AudioEnginePropertyID>{self.id}</AudioEnginePropertyID>
  <Restrictions>
    <ValueRestriction>
      <Enumeration Type="{self.xmlTypeName}"> 
        {options}
      </Enumeration>
    </ValueRestriction>
  </Restrictions>{_generate_dependencies()}
</Property>'''.splitlines()

        def _generate_float_gui_lines():
            user_interface = self.userInterface or f'Step="0.1" Fine="0.001" Decimals="3" UIMax="{self.maxValue}"'
            return f'''<Property Name="{self.propertyName}" Type="{self.xmlTypeName}" {support_rtpc_type} {data_meaning} DisplayName="{self.displayName}">
  <UserInterface {user_interface} />
  <DefaultValue>{self.defaultValue}</DefaultValue>
  <AudioEnginePropertyID>{self.id}</AudioEnginePropertyID>
  <Restrictions>
    <ValueRestriction>
      <Range Type="{self.xmlTypeName}">
        <Min>{self.minValue}</Min>
        <Max>{self.maxValue}</Max>
      </Range>
    </ValueRestriction>
  </Restrictions>{_generate_dependencies()}
</Property>'''.splitlines()

        if self.type_ == 'bool':
            return _generate_bool_gui_lines()

        if self.type_ == 'int' or self.type_ == 'uint':
            return _generate_int_gui_lines()

        if self.type_ == 'float':
            return _generate_float_gui_lines()

        raise NotImplementedError(f'Parameter type "{self.type_}" not supported.')

    def dump_parameter_doc(self, docs_dir: str):
        for lang in self.description:
            output_path = osp.join(docs_dir, f'{lang["language"]}', f'{self.propertyName}.md')
            doc_str = f'''##{self.displayName}

{lang['text']}

Range: {self.minValue} - {self.maxValue} <br/>'''
            util.save_text(output_path, doc_str)

    def generate_win32_controls(self) -> list[str]:
        row_height = 18
        vertical_pos = row_height * self.id + 6
        title = f'LTEXT "{self.displayName}",IDC_STATIC,0,{vertical_pos + 2},48,10'
        control_pos = f'48,{vertical_pos},64,12'
        if self.type_ == 'bool':
            control = f'CONTROL "{self.propertyName}",IDC_{self.propertyName},"Button",BS_AUTOCHECKBOX | WS_TABSTOP,' + control_pos
        elif self.type_ == 'int':
            if self.enumeration:
                options = ', '.join([f'{opt["value"]}:{opt["displayName"]}' for opt in self.enumeration])
                control = f'LTEXT "Class=Combo;Prop={self.propertyName};Options={options}",IDC_{self.propertyName},' + control_pos
            else:
                control = f'LTEXT "Class=Spinner;Prop={self.propertyName};Min={self.minValue};Max={self.maxValue}",IDC_{self.propertyName},' + control_pos
        elif self.type_ == 'float':
            control = f'LTEXT "Class=SuperRange;Prop={self.propertyName}",IDC_{self.propertyName},' + control_pos
        else:
            raise ValueError(f'Unknown type: {self.type_}')
        return [title, control]


class ParameterGenerator:
    def __init__(self, path_man, is_forced=False, generate_gui_resource=False):
        self.pathMan = path_man
        self.isForced = is_forced
        self.generateGuiResource = generate_gui_resource
        self.innerTypes: dict[str, InnerType] = {}
        self.parameters: dict[str, Parameter] = {}
        self.pluginInfo: Optional[PluginInfo] = None

    def main(self):
        self.load_parameter_config()
        self._generate()

    def load_parameter_config(self):
        proj_config = ProjectConfig(self.pathMan)

        # load plugin info
        self.pluginInfo = proj_config.plugin_info()

        # load parameters
        if not proj_config.has_parameters():
            return
        for name, define in proj_config.parameter_defines().items():
            self.parameters[name] = Parameter.create(name, define)
        for instance in proj_config.parameter_from_templates():
            self.__load_with_template(instance, proj_config.parameter_templates())
        self.innerTypes = {name: InnerType.create(name, dict_define) for name, dict_define in proj_config.parameter_inner_types().items()}
        for instance in proj_config.parameter_from_inner_types():
            self.__load_with_inner_type(instance)

        for i, param in enumerate(self.parameters.values()):
            param.assign_id(i)
            param.generate_names()
            # link parameter dependency
            for dep in param.dependencies:
                dep['obj'] = self.parameters[dep['name']]

    def _generate(self):
        def _generate_fx_params_h():
            target = 'SoundEnginePlugin/ProjectNameFXParams.h'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_ids(), dst, '// [ParameterID]', '// [/ParameterID]')
            util.substitute_lines_in_file(self.__generate_inner_types(), dst, '// [InnerTypes]', '// [/InnerTypes]')
            util.substitute_lines_in_file(self.__generate_declarations(struct='InnerType'), dst, '// [InnerTypeDeclaration]',
                                          '// [/InnerTypeDeclaration]')
            util.substitute_lines_in_file(self.__generate_declarations(struct='RTPC'), dst, '// [RTPCDeclaration]',
                                          '// [/RTPCDeclaration]')
            util.substitute_lines_in_file(self.__generate_declarations(struct='NonRTPC'), dst, '// [NonRTPCDeclaration]',
                                          '// [/NonRTPCDeclaration]')

        def _generate_fx_params_cpp():
            target = 'SoundEnginePlugin/ProjectNameFXParams.cpp'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_init(), dst, '// [ParameterInitialization]',
                                          '// [/ParameterInitialization]')
            util.substitute_lines_in_file(self.__generate_read_bank_data(), dst, '// [ReadBankData]',
                                          '// [/ReadBankData]')
            util.substitute_lines_in_file(self.__generate_set_parameter(), dst, '// [SetParameters]',
                                          '// [/SetParameters]', withindent=False)

        def _generate_wwise_plugin_h():
            target = 'WwisePlugin/ProjectNamePlugin.h'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_property_name_declaration(), dst, '// [PropertyNames]',
                                          '// [/PropertyNames]')

        def _generate_wwise_plugin_cpp():
            target = 'WwisePlugin/ProjectNamePlugin.cpp'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_property_name_definition(), dst, '// [PropertyNames]',
                                          '// [/PropertyNames]')
            util.substitute_lines_in_file(self.__generate_write_bank_data(), dst, '// [WriteBankData]',
                                          '// [/WriteBankData]')

        def _generate_wwise_xml():
            target = 'WwisePlugin/ProjectName.xml'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_parameter_gui(), dst, '<!-- [ParameterGui] -->',
                                          '<!-- [/ParameterGui] -->')
            util.substitute_lines_in_file(self.__generate_platform_support(), dst, '<!-- [PlatformSupport] -->',
                                          '<!-- [/PlatformSupport] -->')
            util.substitute_lines_in_file(self.__generate_plugin_info(), dst, '<!-- [PluginInfo] -->',
                                          '<!-- [/PluginInfo] -->')

        def _generate_doc():
            for param in self.parameters.values():
                param.dump_parameter_doc(self.pathMan.docsDir)

        def _generate_win32_gui_resource():
            target = 'WwisePlugin/ProjectName.rc'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_win32_controls(), dst, '// [Controls]', '// [/Controls]')
            target = 'WwisePlugin/resource.h'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_win32_idc(), dst, '// [IDC]', '// [/IDC]')
            target = 'WwisePlugin/Win32/ProjectNamePluginGUI.cpp'
            dst = copy_template(target, self.pathMan, self.isForced)
            util.substitute_lines_in_file(self.__generate_win32_property_table(), dst, '// [PropertyTable]', '// [/PropertyTable]')

        _generate_fx_params_h()
        _generate_fx_params_cpp()
        _generate_wwise_plugin_h()
        _generate_wwise_plugin_cpp()
        _generate_wwise_xml()
        _generate_doc()
        if self.generateGuiResource:
            _generate_win32_gui_resource()

    def __generate_ids(self):
        lines = []
        for i, param in enumerate(self.parameters.values()):
            lines.append(param.generate_param_id())
        lines.append(f'static constexpr AkUInt32 NUM_PARAMS = {len(self.parameters)};')
        return auto_add_line_end(lines)

    def __generate_inner_types(self):
        lines = []
        for i, inner_type in enumerate(self.innerTypes.values()):
            lines.append(inner_type.generate_struct_defines())
        return auto_add_line_end(lines)

    def __load_with_template(self, instance, templates):
        template = copy.deepcopy(templates[instance['template']])
        for key, value in instance.get('override', {}).items():
            template[key] = value
        for dep in template.get('dependencies', []):
            dep['name'] = dep['name'] % {'suffix': instance['suffix']}

        name = f"{instance['template']}_{instance['suffix']}"
        self.parameters[name] = Parameter.create(name, template)

    def __load_with_inner_type(self, define):
        inner_type = self.innerTypes[define['inner_type']]
        overrides = define.get('overrides', {})

        for type_field in inner_type.fields:
            field_name = type_field.name
            instance = copy.deepcopy(type_field)
            if field_name in overrides:
                instance.defaultValue = overrides[field_name]
            for dep in instance.dependencies:
                dep['name'] = f"{inner_type.name}_{dep['name'] % {'suffix': define['suffix']}}"

            instance.name = f"{inner_type.name}_{field_name}_{define['suffix']}"
            instance.basename = field_name
            instance.parent = inner_type
            instance.suffix = define['suffix']
            self.parameters[instance.name] = instance

    def __generate_declarations(self, struct):
        lines = []
        for param in self.parameters.values():
            if param.struct == struct:
                lines.append(param.generate_declaration())
        # remove duplicate lines(when inner type is used)
        lines = list(OrderedDict.fromkeys(lines).keys())
        return auto_add_line_end(lines)

    def __generate_init(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_init())
        return auto_add_line_end(lines)

    def __generate_read_bank_data(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_read_bank_data())
        return auto_add_line_end(lines)

    def __generate_set_parameter(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_set_parameter())
        return auto_add_line_end(lines)

    def __generate_property_name_declaration(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_property_name_declaration())
        return auto_add_line_end(lines)

    def __generate_property_name_definition(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_property_name_definition())
        return auto_add_line_end(lines)

    def __generate_write_bank_data(self):
        lines = []
        for param in self.parameters.values():
            lines.append(param.generate_write_bank_data())
        return auto_add_line_end(lines)

    def __generate_parameter_gui(self):
        lines = []
        for param in self.parameters.values():
            lines.extend(param.generate_parameter_gui())
        return auto_add_line_end(lines)

    def __generate_platform_support(self):
        logging.warning('Deprecated tag: "PlatformSupport", please migrate to "PluginInfo" instead.')
        return auto_add_line_end(self.pluginInfo.generate_platform_support())

    def __generate_plugin_info(self):
        return auto_add_line_end(self.pluginInfo.generate_plugin_info())

    def __generate_win32_controls(self):
        lines = []
        for param in self.parameters.values():
            lines.extend(param.generate_win32_controls())
        return auto_add_line_end(lines)

    def __generate_win32_idc(self):
        lines = []
        for i, param in enumerate(self.parameters.values()):
            lines.append(f'#define IDC_{param.propertyName} {i + 1001}')
        return auto_add_line_end(lines)

    def __generate_win32_property_table(self):
        lines = []
        for param in self.parameters.values():
            if param.type_ != 'bool':
                continue
            lines.append(f'    AK_WWISE_PLUGIN_GUI_WINDOWS_POP_ITEM(IDC_{param.propertyName}, sz{param.propertyName})')
        return auto_add_line_end(lines)



import logging
import os
import re
import shutil
import subprocess
import tomllib
import os.path as osp
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from pathlib import Path

import kkpyutil as util

from wpe.pathman import PathMan


def overwrite_copy(src: str, dst: str):
    if osp.isfile(src):
        copy_file(src, dst)

    if osp.isdir(src):
        copy_tree(src, dst)


def path_is_under(child: str, parent: str) -> bool:
    return Path(parent) in Path(child).parents


def load_toml(path):
    with open(path, 'rb') as f:
        toml_content = tomllib.load(f)
    return toml_content


def git_mv(src, dst):
    try:
        util.run_cmd(['git', 'mv', src, dst])
    except subprocess.CalledProcessError:
        # git ignored files will raise error, fallback to os.rename
        os.rename(src, dst)
    except Exception as e:
        logging.error(f'Error moving {src} to {dst} with `git mv`: {e}')
        raise e


def replace_in_basename(path: str, old: str, new: str, count=-1):
    return osp.join(osp.dirname(path), osp.basename(path).replace(old, new, count))


def remove_path(path, ignore_errors=True):
    if osp.isfile(path):
        os.remove(path)
    if osp.isdir(path):
        shutil.rmtree(path, ignore_errors=ignore_errors)


def replace_content_in_file(file_path, org, dst):
    content = util.load_text(file_path)
    util.save_text(file_path, content.replace(org, dst))


def path_is_under_git_repo(path: str) -> bool:
    git_exe = shutil.which('git')
    if not git_exe:
        return False
    try:
        res = util.run_cmd([git_exe, '-C', path, 'rev-parse', '--is-inside-work-tree'])
        return res.stdout.decode().strip() == 'true'
    except Exception as e:
        logging.error(f'Error checking git repo: {e}')
        return False


def remove_ansi_color(text):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)


def auto_add_line_end(lines: list[str]):
    for i, line in enumerate(lines):
        if not line.endswith('\n'):
            lines[i] += '\n'
    return lines


def add_indent(lines: list[str], indent: int):
    return [f'{" " * indent}{line}' for line in lines]


def copy_template(relative, pathman: PathMan, is_forced=False):
    def _need_overwrite(_dst):
        if is_forced:
            return True
        if osp.isfile(_dst):
            content = util.load_text(_dst)
            return '[wp-enhanced template]' not in content
        return True

    src = osp.join(pathman.templatesDir, relative)
    dst = replace_in_basename(osp.join(pathman.root, relative), 'ProjectName', pathman.pluginName)
    if not _need_overwrite(dst):
        logging.info(f'Skip copying template "{osp.basename(src)}". Use -f to force overwrite.')
        return dst
    if osp.isfile(src):
        overwrite_copy(src, dst)
        util.substitute_keywords_in_file(dst,
                                         {
                                             'name': pathman.pluginName,
                                             'display_name': pathman.pluginName,
                                             'plugin_id': pathman.pluginId,
                                         })
        return dst
    else:
        raise FileNotFoundError(f'File not found: {src}')



import logging
import os.path as osp
import platform
import subprocess
import glob
from typing import Optional

# 3rd party
import kkpyutil as util

# project
import wpe.util as wpe_util
from wpe.pathman import PathMan
from wpe.wp_wrapper import WpWrapper
from wpe.parameter import ParameterGenerator
from wpe.hook_processor import HookProcessor
from wpe.project_config import ProjectConfig, PlatformTarget
from wpe.test_runner import TestRunner
from wpe.renamer import Renamer
from wpe import constants


class Worker:
    def __init__(self, args, path_man=None):
        self.args = args

        self.wpWrapper = WpWrapper()

        # lazy inits
        self.pathMan: Optional[PathMan] = path_man
        self.projConfig: Optional[ProjectConfig] = None
        self.targetPlatforms: list[PlatformTarget] = []

        self.wwiseProcName = ''
        self.terminatedWwise = False

    @staticmethod
    def create_platform(args):
        system = platform.system()
        if system == 'Windows':
            return WindowsWorker(args)
        if system == 'Darwin':
            return MacWorker(args)
        raise NotImplementedError(f'Not implemented for this platform: {system}')

    def _lazy_load_configs(self):
        self.pathMan = self.pathMan or PathMan()
        self.projConfig = ProjectConfig(self.pathMan)
        self.targetPlatforms = self.projConfig.target_platforms()
        if self.args.platforms:
            self.targetPlatforms = [plt for plt in self.targetPlatforms if plt.platform in self.args.platforms]
        HookProcessor().init(self.pathMan, self.args.configuration, self.args.withHooks)

    def main(self):
        self.wpWrapper.validate_env()

        if self.args.wp:
            return self.wp()

        if self.args.new:
            return self.new()

        self._lazy_load_configs()

        if self.args.initWpe:
            return self.init_wpe()

        if self.args.premake:
            self.premake()

        if self.args.generateParameters:
            self.generate_parameters()

        if self.args.build:
            self.build()

        if self.args.test:
            self.test()

        if self.args.pack:
            self.pack()

        if self.args.fullPack:
            self.full_pack()

        if self.args.bump:
            self.bump()

        if self.args.rename:
            self.rename()

    def wp(self):
        logging.info('Run wp.py')
        self.wpWrapper.wp(self.args.wp)

    def new(self):
        def _copy_win32_gui_resource():
            wpe_util.copy_template('WwisePlugin/ProjectName.rc', self.pathMan)
            wpe_util.copy_template('WwisePlugin/resource.h', self.pathMan)
            wpe_util.copy_template('WwisePlugin/Win32/ProjectNamePluginGUI.h', self.pathMan)
            wpe_util.copy_template('WwisePlugin/Win32/ProjectNamePluginGUI.cpp', self.pathMan)

        logging.info('Create new project')
        self.wpWrapper.new()
        self.init_wpe()
        self._lazy_load_configs()
        _copy_win32_gui_resource()
        self.premake()
        logging.info('Next step: implement your plugin, build with hooks by command: wpe -b -H')

    def init_wpe(self):
        def _append_to_gitignore():
            gitignore = osp.join(self.pathMan.root, '.gitignore')
            if osp.exists(gitignore):
                with open(gitignore, 'a') as f:
                    f.write(constants.extra_gitignore)
        logging.info('Initialize wpe project config')
        self.pathMan = self.pathMan or PathMan()
        HookProcessor().init(self.pathMan, self.args.configuration, self.args.withHooks)
        _append_to_gitignore()
        wpe_util.overwrite_copy(
            osp.join(self.pathMan.templatesDir, '.wpe'),
            osp.join(self.pathMan.configDir)
        )

    @HookProcessor().register('premake')
    def premake(self):
        logging.info('Premake project')
        platforms = set([plt.platform for plt in self.targetPlatforms])
        for plt in platforms:
            self.wpWrapper.premake(plt)

    @HookProcessor().register('generate_parameters')
    def generate_parameters(self):
        def clear_existing_doc():
            util.remove_tree(self.pathMan.docsDir)
            util.remove_tree(self.pathMan.htmlDocsDir)
        clear_existing_doc()
        parameter_manager = ParameterGenerator(self.pathMan,
                                               is_forced=self.args.force,
                                               generate_gui_resource=self.args.gui)
        parameter_manager.main()
        self.wpWrapper.build('Documentation')

    @HookProcessor().register('build')
    def build(self):
        self._terminate_wwise()
        self._build()
        self._reopen_wwise()

    @HookProcessor().register('test')
    def test(self):
        # self.args.configuration = 'Release'
        # self.targetPlatforms = [PlatformTarget({
        #     'platform': 'Windows_vc160',
        #     'architectures': ['x64'],
        # })]
        # self.build()
        TestRunner.create_platform(self.pathMan).main()

    @HookProcessor().register('pack')
    def pack(self):
        def _collect_packages(_output_dir):
            util.remove_tree(_output_dir)
            for pkg in glob.iglob(osp.join(self.pathMan.root, f'{self.pathMan.pluginName}*.tar.xz')):
                util.move_file(pkg, _output_dir, isdstdir=True)
            util.move_file(osp.join(self.pathMan.root, 'bundle.json'), _output_dir, isdstdir=True)

        def _zip_bundle(_output_dir):
            util.zip_dir(_output_dir)

        logging.info('Package plugin and generate bundle')
        version_code, build_number = self.wpWrapper.wwiseVersion.rsplit('.', 1)
        build_number = self.projConfig.version()

        plugin_version = f'{version_code}.{build_number}'
        output_dir = osp.join(self.pathMan.distDir, f'{self.pathMan.pluginName}_v{version_code}_Build{build_number}')
        self.wpWrapper.package('Common', '-v', plugin_version)
        self.wpWrapper.package('Documentation', '-v', plugin_version)
        for plt in self.projConfig.all_platforms():
            self.wpWrapper.package(plt, '-v', plugin_version)
        self.wpWrapper.generate_bundle('-v', plugin_version)
        _collect_packages(output_dir)
        _zip_bundle(output_dir)
        logging.info(f'Saved to {output_dir}')

    def full_pack(self):
        HookProcessor().process_pre_hook('build')
        for plt in self.targetPlatforms:
            args = [plt.platform, '-c', 'Release', '-x'] + plt.architectures
            if plt.need_toolset():
                args.extend(['-t', plt.toolset()])
            self.wpWrapper.build(*args)
            if plt.is_authoring():
                continue
            args[2] = 'Profile'
            self.wpWrapper.build(*args)
        HookProcessor().process_post_hook('build')
        self.pack()

    @HookProcessor().register('bump')
    def bump(self):
        logging.info('Bump wpe project version')
        self.projConfig.bump()
        logging.info(f'Version bumped to {self.projConfig.version()}')

    @HookProcessor().register('rename')
    def rename(self):
        logging.info(f'Rename plugin from {self.pathMan.pluginName} to {self.args.rename}.')
        res = input('Commit your changes before renaming is recommended. Continue? [y/n]') == 'y'
        if not res:
            return
        Renamer(self.args.rename, self.pathMan, self.projConfig).main()
        self.premake()
        logging.info('Rename completed, check your changes with git status.')

    def _build(self):
        logging.info('Build plugin')
        for plt in self.targetPlatforms:
            args = [plt.platform, '-c', self.args.configuration, '-x'] + plt.architectures
            if plt.need_toolset():
                args.extend(['-t', plt.toolset()])
            self.wpWrapper.build(*args)
        self.wpWrapper.build('Documentation')

    def _terminate_wwise(self):
        if self.args.force:
            try:
                util.kill_process_by_name(self.wwiseProcName, forcekill=True)
                self.terminatedWwise = True
            except subprocess.CalledProcessError:
                pass

    def _reopen_wwise(self):
        raise NotImplementedError('subclass it')


class WindowsWorker(Worker):
    def __init__(self, args):
        super().__init__(args)
        self.wwiseProcName = 'Wwise.exe'

    def _reopen_wwise(self):
        if self.args.force and self.terminatedWwise:
            wwise_exe = osp.join(self.wpWrapper.wwiseRoot, 'Authoring/x64/Release/bin/Wwise.exe')
            util.run_daemon([wwise_exe])


class MacWorker(Worker):
    def __init__(self, args):
        super().__init__(args)
        self.wwiseProcName = 'wine64-preloader'

    def _reopen_wwise(self):
        if self.args.force and self.terminatedWwise:
            wwise_app = osp.join(self.wpWrapper.wwiseRoot, 'Wwise.app')
            util.run_daemon(['open', wwise_app])



import logging
import os.path as osp
import platform

import kkpyutil as util
import requests

from wpe.pathman import PathMan
from wpe.util import overwrite_copy, remove_ansi_color


class TestRunner:
    def __init__(self, path_man: PathMan):
        self.pathMan = path_man

    @staticmethod
    def create_platform(path_man: PathMan):
        system = platform.system()
        if system == 'Windows':
            return _WindowsTestRunner(path_man)
        raise NotImplementedError(f'Not implemented for this platform: {system}')

    def main(self):
        self._lazy_create_test_project()
        self._build_test_project()
        self._run_test()

    def _lazy_create_test_project(self):
        def lazy_download_catch2_src():
            catch2_files = ['catch_amalgamated.hpp', 'catch_amalgamated.cpp']
            catch2_dir = osp.join(self.pathMan.testDir, 'catch2')
            if all(osp.isfile(osp.join(catch2_dir, f)) for f in catch2_files):
                return
            logging.info('downloading catch2 source files')
            download_base_url = 'https://github.com/catchorg/Catch2/releases/download/v3.5.3/'
            for file in catch2_files:
                res = requests.get(download_base_url + file)
                util.save_text(osp.join(catch2_dir, file), res.text)

        def lazy_copy_test_src():
            test_src_files = ['CMakeLists.txt', 'main.cpp', '.gitignore', 'util']
            if all(osp.exists(osp.join(self.pathMan.testDir, p)) for p in test_src_files):
                return
            logging.info('copying test source files')
            test_templates_dir = osp.join(self.pathMan.templatesDir, 'test')
            for file in test_src_files:
                overwrite_copy(osp.join(test_templates_dir, file), osp.join(self.pathMan.testDir, file))

            util.substitute_keywords_in_file(osp.join(self.pathMan.testDir, 'CMakeLists.txt'), {
                'name': self.pathMan.pluginName,
                'test_util': self.pathMan.testUtilDir.replace('\\', '/')
            })

        def extract_includes_from_premake():
            from lupa import LuaRuntime
            lua = LuaRuntime(unpack_returned_tuples=True)
            lua.globals()['_AK_PREMAKE'] = True
            plugin_table = lua.execute(util.load_text(self.pathMan.premakePluginLua))
            includes = list(plugin_table['sdk']['static']['includedirs'].values())
            return includes

        def sync_includes_from_premake():
            includes = extract_includes_from_premake()
            inserts = [f'include_directories({include})\n' for include in includes]
            cmakelists_file = osp.join(self.pathMan.testDir, 'CMakeLists.txt')
            util.substitute_lines_in_file(inserts, cmakelists_file,
                                          '# [PremakeDefinedIncludes]',
                                          '# [/PremakeDefinedIncludes]')

        lazy_download_catch2_src()
        lazy_copy_test_src()
        sync_includes_from_premake()

    def _build_test_project(self):
        util.run_cmd(
            [
                'cmake',
                '.',
                '-B',
                'build'
            ],
            cwd=self.pathMan.testDir
        )

    def _run_test(self):
        raise NotImplementedError('subclass it')


class _WindowsTestRunner(TestRunner):
    def _build_test_project(self):
        super()._build_test_project()
        util.run_cmd(
            [
                'msbuild',
                'build\\test.sln',
                '/verbosity:quiet',
                '/property:Configuration=RelWithDebInfo',
            ],
            cwd=self.pathMan.testDir
        )

    def _run_test(self):
        divider = '-------------------------------------------------------------------------------'
        log_lines = [
            divider,
            'CPU Info'
        ]

        query_cpu_info_proc = util.run_cmd(['wmic', 'CPU', 'GET', 'name'])
        log_lines.extend(filter(None, query_cpu_info_proc.stdout.decode(util.LOCALE_CODEC).splitlines()))
        log_lines.append(divider)

        test_proc = util.run_cmd(
            [
                osp.join(self.pathMan.testDir, 'build\\RelWithDebInfo\\test.exe'),
                '--colour-mode', 'ansi',
                '--benchmark-samples', '10',
                '-s'
            ],
            cwd=self.pathMan.testDir
        )
        log_lines.extend(remove_ansi_color(test_proc.stdout.decode(util.LOCALE_CODEC)).splitlines())

        util.save_lines(osp.join(self.pathMan.testDir, 'test_benchmark_report.txt'), log_lines, addlineend=True)



import subprocess
import os.path as osp


def fetch_git_repo(url, dst_dir, branch='master'):
    if osp.exists(dst_dir):
        print(f'Updating {dst_dir}')
        subprocess.check_call(['git', 'pull'], cwd=dst_dir)
        return
    print(f'Cloning {url} to {dst_dir}')
    subprocess.check_call(['git', 'clone', '-b', branch, url, dst_dir])


def main(**kwargs):
    external_repo_dir = osp.join(kwargs['proj_root'], 'external')
    # add external repo fetching here
    # fetch_git_repo('${REPO_URL}', osp.join(external_repo_dir, '&{REPO_NAME}'))



import glob
import os
import os.path as osp
import subprocess

from wpe.pathman import PathMan
from wpe.project_config import ProjectConfig
import wpe.util as wutil


class Renamer:
    def __init__(self, new_name: str, pathman: PathMan, proj_config: ProjectConfig):
        self.oldName = pathman.pluginName
        self.newName = new_name
        self.pathMan = pathman
        self.projConfig = proj_config
        self.renameMethod = wutil.git_mv if wutil.path_is_under_git_repo(self.pathMan.root) else os.rename

    def main(self):
        self._delete_generated_files()
        self._migrate_sources()
        self._move_proj_folder()

    def _delete_generated_files(self):
        targets = [
            f'{self.oldName}*.xcworkspace',
            f'{self.oldName}*.mk',
            f'{self.oldName}*.sln',
            f'*/{self.oldName}*.aps',
            f'*/{self.oldName}*.xcodeproj',
            f'*/{self.oldName}*.mk',
            f'*/{self.oldName}*.vcxproj*',
        ]
        for target in targets:
            for path in glob.iglob(osp.join(self.pathMan.root, target)):
                wutil.remove_path(path)

    def _migrate_sources(self):
        targets = [
            'FactoryAssets/Manifest.xml',
            f'SoundEnginePlugin/{self.oldName}FX*',
            f'WwisePlugin/{self.oldName}*',
            f'WwisePlugin/Win32/{self.oldName}*',
            'additional_artifacts.json',
            'bundle_template.json',
            f'{self.oldName}Config.h',
            'PremakePlugin.lua',
        ]
        for target in targets:
            for old_path in glob.iglob(osp.join(self.pathMan.root, target)):
                if not osp.exists(old_path):
                    continue
                if self.oldName in osp.basename(old_path):
                    new_path = wutil.replace_in_basename(osp.join(self.pathMan.root, old_path), self.oldName, self.newName, count=1)
                    self.renameMethod(old_path, new_path)
                else:
                    new_path = old_path
                wutil.replace_content_in_file(new_path, self.oldName, self.newName)

    def _move_proj_folder(self):
        new_proj_root = osp.join(osp.dirname(self.pathMan.root), self.newName)
        os.makedirs(new_proj_root, exist_ok=True)
        for item in os.listdir(self.pathMan.root):
            self.renameMethod(osp.join(self.pathMan.root, item), osp.join(new_proj_root, item))
        self.pathMan.refresh_paths(new_proj_root)



import os
import os.path as osp
import platform

import kkpyutil as util


def create_sdk_symlink(wwise_sdk):
    temp_sdk_dir = 'C:\\temp\\wpe\\WWISESDK' if platform.system() == 'Windows' else osp.expanduser(
        '~/temp/wpe/WWISESDK')
    if osp.exists(temp_sdk_dir):
        if osp.islink(temp_sdk_dir):
            os.remove(temp_sdk_dir)
        else:
            raise FileExistsError(f'{temp_sdk_dir} exists and is not a symlink to WWISESDK')
    os.makedirs(osp.dirname(temp_sdk_dir), exist_ok=True)
    os.symlink(wwise_sdk, temp_sdk_dir)
    return temp_sdk_dir


def inject_wwise_sdk_for_android(func):
    def wrapper(*args, **kwargs):
        plt = args[0] if args else ''
        if plt != 'Android':
            return func(*args, **kwargs)
        org_sdk_dir = os.getenv('WWISESDK')
        temp_sdk_dir = create_sdk_symlink(org_sdk_dir)
        os.environ['WWISESDK'] = temp_sdk_dir
        res = func(*args, **kwargs)
        os.environ['WWISESDK'] = org_sdk_dir
        return res
    return wrapper


class WpWrapper:
    def __init__(self):
        self.wwiseRoot: str = os.getenv('WWISEROOT')
        self.wwiseSDKRoot: str = os.getenv('WWISESDK')
        self.wwiseVersion: str = self._load_wwise_version()
        self.wpScriptDir = osp.join(self.wwiseRoot, 'Scripts/Build/Plugins')
        util.lazy_prepend_sys_path([self.wpScriptDir])

        self.subcommands = (
            'build',
            'generate_bundle',
            'new',
            'package',
            'premake',
        )

    def _load_wwise_version(self) -> str:
        install_entry = util.load_json(osp.join(self.wwiseRoot, 'install-entry.json'))
        version = install_entry['bundle']['version']
        return f'{version["year"]}.{version["major"]}.{version["minor"]}.{version["build"]}'

    def validate_env(self):
        if self.wwiseRoot is None:
            raise EnvironmentError(f'Unknown env variable: WWISEROOT\n  - Try setting environment variables in Wwise '
                                   f'Launcher')

        if self.wwiseSDKRoot is None:
            raise EnvironmentError(f'Unknown env variable: WWISESDK\n  - Try setting environment variables in Wwise '
                                   f'Launcher')

        if not osp.isfile((wp := osp.join(self.wpScriptDir, 'wp.py'))):
            raise FileNotFoundError(f'"{wp}" not found.')

        for subcommand in self.subcommands:
            if not osp.isfile((subcommand_file := osp.join(self.wpScriptDir, f'{subcommand}.py'))):
                raise FileNotFoundError(f'Subcommand "{subcommand_file}" not found.')

    def wp(self, args):
        subcommand = args[0]
        if subcommand not in self.subcommands:
            raise ValueError(f'Unknown subcommand: {subcommand}')
        return self.run(subcommand, *args[1:])

    @staticmethod
    @inject_wwise_sdk_for_android
    def build(*args):
        import wpe.wp_patch.build as wpe_build
        res = wpe_build.run(args)
        if res != 0:
            raise RuntimeError(f'Build failed. Exit code: {res}')
        return res

    def generate_bundle(self, *args):
        return self.run('generate_bundle', *args)

    def new(self, *args):
        return self.run('new', *args)

    @staticmethod
    def package(*args):
        import wpe.wp_patch.package as wpe_package
        res = wpe_package.run(args)
        return res

    @staticmethod
    @inject_wwise_sdk_for_android
    def premake(*args):
        import wpe.wp_patch.premake as wpe_premake
        res = wpe_premake.run(args)
        return res

    def run(self, subcommand, *args):
        subcommand_module = util.safe_import_module(osp.basename(subcommand), self.wpScriptDir)
        return subcommand_module.run(args)
